// Lib :
#include <Lib/Printer.h>
#include <Lib/Out.h>
#include <Lib/Arguments.h>
#include <Lib/fmanip.h>
#include <Lib/Debug.h>
#include <Lib/dirmanip.h>
#include <Lib/FileWriter.h>
#include <Lib/sout.h>

// this :
#include <obuild/Builder.h>
#include <obuild/Shell.h>

static bool do_bat(Lib::Arguments&); 
static bool do_sh(Lib::Arguments&); 

//////////////////////////////////////////////////////////////////////////////
int main(
 int aArgc
,char** aArgv
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  Lib::Printer printer;
  //Lib::Debug::checkByClass(true);

  bool status = true;

 {Lib::Arguments args(aArgc,aArgv);

  if(args.isAnArgument("-h")) {

    Lib::Out out(printer);
    out << "OS> obuild [options] <file>" << Lib::endl;
    out << " -h : help" << Lib::endl;
    out << " -def : produce a default this.obuild file." << Lib::endl;
    out << " -q : infos about what to build." << Lib::endl;
    out << " -q=[file,uses,libs,dlls,apps,refs] : infos by components." 
        << Lib::endl;
    out << " -sh : produce sh files for UNIX." << Lib::endl;
    out << " -bat : produce bat files for DOS." << Lib::endl;
    out << " -e : erase sh, bat,... files." << Lib::endl;
    out << " -shell : go under the obuild shell.." << Lib::endl;
    out << " -oshell : produce the packages.oshell file." << Lib::endl;
    out << " -shell_from_oshell : go under the obuild shell by taking the package list from the packages.oshell file." << Lib::endl;

  } else if(args.isAnArgument("-q")) {

    std::string what;
    args.find("-q",what);

    std::string file;
    args.file(file);
    if(file=="") file = "this.obuild";

    if(!Lib::fmanip::exists(file)) {
      Lib::Out out(printer);
      out << "obuild :" << Lib::endl;
      out << " No obuild file given and no this.obuild file found." 
          << Lib::endl;
      out << " Use the -def option to produce a default this.obuild file."
          << Lib::endl;
      status = false;
    } else {
      obuild::Builder builder(printer);
      status = builder.query(file,what);
    }

  } else if(args.isAnArgument("-def")) {

    std::string package;
    std::string version;
    if(!obuild::Builder::package_and_version(package,version)) {
      Lib::Out out(printer);
      out << "obuild :"
          << " Can't get package and version from current directory." 
          << Lib::endl;
      status = false;
    } else {
      std::string file = "this.obuild";
      obuild::Builder::out_default_obuild(printer,package,version,file);
    }

  } else if(args.isAnArgument("-shell")) {

    std::string file;
    args.file(file);
    if(file=="") file = "this.obuild";

    std::vector<std::string> reps;
    std::vector<std::string> pvs;
    status = obuild::Builder::findProjectPackages(printer,file,reps,pvs);
    if(status) {
      obuild::Shell shell(printer);
      status = shell.setup(reps,pvs);
      if(status) status = shell.start();
    }

  } else if(args.isAnArgument("-oshell")) { 
    // Produce package.oshell which is used 
    // to checkout with : obuild -shell_from_oshell.

    std::string file;
    args.file(file);
    if(file=="") file = "this.obuild";

    std::vector<std::string> reps;
    std::vector<std::string> pvs;
    status = obuild::Builder::findProjectPackages(printer,file,reps,pvs);
    if(status) {
      Lib::FileWriter fp("packages.oshell",printer);
      if(!fp.enabled()) {
        Lib::Out out(printer);
        out << "obuild :"
            << " can't open file " << Lib::sout("packages.oshell")
            << Lib::endl;
        status = false;
      } else {
        Lib::Out out(fp);
        unsigned int number = reps.size();
        for(unsigned int index=0;index<number;index++) {
          // If changing the " " separator, change Shell::setupFromFile.
          out << reps[index] << " " << pvs[index] << Lib::endl;
        }
        fp.disable();
      }
    } 

  } else if(args.isAnArgument("-shell_from_oshell")) {
    // Start a shell but by getting packages from packages.oshell.
    // For example used to checkout with :
    //   obuild -shell_from_oshell
    //   obuild>> co

    std::string file;
    if(!args.file(file)) file = "packages.oshell";

    obuild::Shell shell(printer);
    status = shell.setupFromFile(file);
    if(status) status = shell.start();

  } else if(args.isAnArgument("-e")) {

    obuild::Builder builder(printer);      
    builder.cleanup();

  } else if(args.isAnArgument("-bat")) {

    status = do_bat(args);

  } else if(args.isAnArgument("-sh")) {

    status = do_sh(args);

  } else if(args.number()==1) {

    status = do_sh(args);
    if(status) status = do_bat(args);

  } else {

    Lib::Out out(printer);
    out << "obuild :" << Lib::endl;
    out << " unknown arguments."
        << Lib::endl;
    status = false;

  }}

  Lib::Debug::balance(printer);

  return status?EXIT_SUCCESS:EXIT_FAILURE;
}
//////////////////////////////////////////////////////////////////////////////
bool do_bat(
 Lib::Arguments& aArgs
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  Lib::Printer printer;

  int verboseLevel = aArgs.isAnArgument("-verbose")?1:0;
 
  std::string file;
  aArgs.file(file);

  if(file=="") file = "this.obuild";

  if(!Lib::fmanip::exists(file)) {
    Lib::Out out(printer);
    out << "obuild :" << Lib::endl;
    out << " No obuild file given and no this.obuild file found." 
        << Lib::endl;
    out << " Use the -def option to produce a default this.obuild file."
        << Lib::endl;
    return false;
  } else {
    obuild::Builder builder(printer);      
    return builder.build(file,obuild::Builder::build_bat);
  }
}
//////////////////////////////////////////////////////////////////////////////
bool do_sh(
 Lib::Arguments& aArgs
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  Lib::Printer printer;

  int verboseLevel = aArgs.isAnArgument("-verbose")?1:0;
 
  std::string file;
  aArgs.file(file);

  if(file=="") file = "this.obuild";

  if(!Lib::fmanip::exists(file)) {
    Lib::Out out(printer);
    out << "obuild :" << Lib::endl;
    out << " No obuild file given and no this.obuild file found." 
        << Lib::endl;
    out << " Use the -def option to produce a default this.obuild file."
        << Lib::endl;
    return false;
  } else {
    obuild::Builder builder(printer);      
    return builder.build(file,obuild::Builder::build_sh);
  }
}
